@@ -39,12 +39,12 @@
 
 G_DEFINE_TYPE (HDPluginLoaderPython, hd_plugin_loader_python, HD_TYPE_PLUGIN_LOADER);
 
-struct _HDPluginLoaderPythonPrivate 
+struct _HDPluginLoaderPythonPrivate
 {
   gboolean initialised;
 };
 
-static void 
+static void
 hd_plugin_loader_python_destroy_plugin (GtkObject *object, gpointer user_data)
 {
   PyObject *pModule, *pObject;
@@ -58,39 +58,6 @@ hd_plugin_loader_python_destroy_plugin (GtkObject *object, gpointer user_data)
   PyGC_Collect();
 }
 
-static int
-hd_plugin_loader_python_check_type (PyObject *pObject)
-{
-  PyObject *module;
-  PyTypeObject *PyHDHomePluginItem_Type;
-  int ret = 0;
-
-  if ((module = PyImport_ImportModule ("hildondesktop")) != NULL)
-  {
-    PyHDHomePluginItem_Type = (PyTypeObject *)PyObject_GetAttrString(module, "HomePluginItem");
-    if (PyHDHomePluginItem_Type == NULL)
-    {
-      PyErr_SetString (PyExc_ImportError, "cannot import name HomePluginItem from hildondesktop");
-    }
-    else if (pygobject_check(pObject, PyHDHomePluginItem_Type))
-    {
-      ret = 1;
-    }
-    else
-    {
-      PyErr_SetString (PyExc_TypeError, "returned object must be a HDHomePluginItem");
-    }
-    Py_XDECREF (PyHDHomePluginItem_Type);
-  }
-  else
-  {
-    PyErr_SetString (PyExc_ImportError, "could not import hildondesktop");
-  }
-  Py_XDECREF (module);
-
-  return ret;
-}
-
 static PyObject *
 hd_plugin_loader_python_import_module (gchar *module_name)
 {
@@ -134,6 +101,66 @@ hd_plugin_loader_python_import_module (gchar *module_name)
   return pModule;
 }
 
+/* Based on gobject.type_register() and gobject.new() implementations from pygobject */
+static GObject *
+hd_plugin_loader_python_create_object (PyObject    *pClass,
+                                       const gchar *plugin_id)
+{
+  GObject *object = NULL;
+
+  PyObject *pModule = PyImport_ImportModule ("gobject");
+  if (pModule != NULL)
+  {
+    PyObject *pDict = PyModule_GetDict (pModule);
+    PyObject *register_func = PyDict_GetItemString (pDict, "type_register");
+    PyObject *new_func = PyDict_GetItemString (pDict, "new");
+    if (register_func != NULL && new_func != NULL)
+    {
+      PyObject *pRes = PyObject_CallFunctionObjArgs (register_func, pClass, NULL);
+      if (pRes != NULL)
+      {
+        Py_DECREF (pRes);
+        PyObject *args = Py_BuildValue ("(O)", pClass);
+        PyObject *kwargs = Py_BuildValue ("{ss}", "plugin_id", plugin_id);
+        PyObject *pObject = PyObject_Call (new_func, args, kwargs);
+        Py_DECREF (args);
+        Py_DECREF (kwargs);
+        if (pObject != NULL)
+        {
+          object = g_object_ref (((PyGObject *) pObject)->obj);
+          g_object_set_data (object, "object", pObject);
+        }
+        else
+        {
+          PyErr_Print ();
+          PyErr_Clear ();
+          g_warning ("Could not create new plugin object");
+        }
+      }
+      else
+      {
+        PyErr_Print ();
+        PyErr_Clear ();
+        g_warning ("Could not register type for plugin object");
+      }
+    }
+    else
+    {
+      PyErr_Print ();
+      PyErr_Clear ();
+      g_warning ("Could not import functions from gobject");
+    }
+  }
+  else
+  {
+    PyErr_Print ();
+    PyErr_Clear ();
+    g_warning ("Could not import gobject");
+  }
+
+  return object;
+}
+
 static GObject *
 hd_plugin_loader_python_open_module (HDPluginLoaderPython  *loader,
                                      const gchar           *plugin_id,
@@ -141,7 +168,7 @@ hd_plugin_loader_python_open_module (HDPluginLoaderPython  *loader,
                                      GError               **error)
 {
   HDPluginLoaderPythonPrivate *priv;
-  PyObject *pModule, *pDict, *pFunc, *pObject;
+  PyObject *pModule, *pDict;
   GObject *object = NULL;
   GError *keyfile_error = NULL;
   gchar *module_file = NULL;
@@ -164,6 +191,7 @@ hd_plugin_loader_python_open_module (HDPluginLoaderPython  *loader,
     return NULL;
   }
 
+  /* Strip directory component from absolute paths */
   if (g_path_is_absolute (module_file))
   {
     module_name = g_path_get_basename (module_file);
@@ -172,12 +200,12 @@ hd_plugin_loader_python_open_module (HDPluginLoaderPython  *loader,
   }
   else
   {
-    module_name = module_file; 
+    module_name = module_file;
   }
 
+  /* Strip extension from module_name */
   if (g_str_has_suffix (module_name, ".py"))
   {
-    /* Strip extension from module_name */
     gchar *tmp = g_strndup (module_name, strlen (module_name) - 3);
     g_free (module_name);
     module_name = tmp;
@@ -188,54 +216,44 @@ hd_plugin_loader_python_open_module (HDPluginLoaderPython  *loader,
   if (pModule != NULL)
   {
     pDict = PyModule_GetDict (pModule);
-
-    pFunc = PyDict_GetItemString (pDict, "hd_plugin_get_object");
-
-    if (pFunc != NULL && PyCallable_Check (pFunc))
+    PyObject *pType = PyDict_GetItemString (pDict, "hd_plugin_type");
+    if (pType != NULL)
     {
-      pObject = PyObject_CallObject (pFunc, NULL);
-
-      if (pObject != NULL)
+      object = hd_plugin_loader_python_create_object (pType, plugin_id);
+      if (object != NULL)
       {
-        if (!hd_plugin_loader_python_check_type (pObject))
+        g_object_set_data (object, "module", pModule);
+        g_signal_connect (G_OBJECT (object),
+                          "destroy",
+                          G_CALLBACK (hd_plugin_loader_python_destroy_plugin),
+                          NULL);
+      }
+      else
+      {
+        Py_DECREF (pModule);
+        if (PyErr_Occurred ())
         {
           PyErr_Print ();
           PyErr_Clear ();
-          Py_DECREF (pObject);
-          Py_DECREF (pModule);
         }
-        else
-        {
-          object = g_object_ref(((PyGObject *) pObject)->obj);
-          g_signal_connect (G_OBJECT (object),
-                            "destroy",
-                            G_CALLBACK (hd_plugin_loader_python_destroy_plugin), 
-                            NULL);
-          g_object_set_data (object, "object", pObject);
-          g_object_set_data (object, "module", pModule);
-        }
-      }
-      else 
-      {
-        Py_DECREF (pModule);
-
-        PyErr_Print ();
-        PyErr_Clear ();
-
-        g_warning ("Failed to call \"hd_plugin_get_object\" in Python module '%s'", module_name);
+        g_set_error (error,
+                     hd_plugin_loader_error_quark (),
+                     HD_PLUGIN_LOADER_ERROR_KEYFILE,
+                     "Could not create plugin object in Python module '%s'", module_name);
       }
     }
-    else 
+    else
     {
       Py_DECREF (pModule);
-
       if (PyErr_Occurred ())
       {
         PyErr_Print ();
         PyErr_Clear ();
       }
-
-      g_warning ("Cannot find \"hd_plugin_get_object\" function in Python module '%s'", module_name);
+      g_set_error (error,
+                   hd_plugin_loader_error_quark (),
+                   HD_PLUGIN_LOADER_ERROR_KEYFILE,
+                   "Could not find \"hd_plugin_type\" in Python module '%s'", module_name);
     }
   }
 
@@ -245,16 +263,16 @@ hd_plugin_loader_python_open_module (HDPluginLoaderPython  *loader,
 }
 
 static GObject *
-hd_plugin_loader_python_load (HDPluginLoader *loader, 
-                              const gchar    *plugin_id, 
-                              GKeyFile       *keyfile, 
+hd_plugin_loader_python_load (HDPluginLoader *loader,
+                              const gchar    *plugin_id,
+                              GKeyFile       *keyfile,
                               GError         **error)
 {
   GObject *object = NULL;
   GError *local_error = NULL;
 
   g_return_val_if_fail (loader, NULL);
- 
+
   if (!keyfile)
   {
     g_set_error (error,
@@ -272,7 +290,7 @@ hd_plugin_loader_python_load (HDPluginLoader *loader,
                                           keyfile,
                                           &local_error);
 
-  if (local_error) 
+  if (local_error)
   {
     g_propagate_error (error, local_error);
 
@@ -294,7 +312,7 @@ hd_plugin_loader_python_finalize (GObject *loader)
 
   if (priv->initialised)
   {
-    Py_Finalize (); 
+    Py_Finalize ();
   }
 
   G_OBJECT_CLASS (hd_plugin_loader_python_parent_class)->finalize (loader);
@@ -309,11 +327,11 @@ hd_plugin_loader_python_init (HDPluginLoaderPython *loader)
 
   Py_Initialize ();
   loader->priv->initialised = TRUE;
-  
+
   init_pygobject ();
   init_pygtk ();
 
-  pStr = PyString_FromString (HD_DESKTOP_MODULE_PATH); 
+  pStr = PyString_FromString (HD_DESKTOP_MODULE_PATH);
   pPath = PySys_GetObject ("path");
   PyList_Append (pPath, pStr);
 }
@@ -326,7 +344,7 @@ hd_plugin_loader_python_class_init (HDPluginLoaderPythonClass *class)
 
   object_class = G_OBJECT_CLASS (class);
   loader_class = HD_PLUGIN_LOADER_CLASS (class);
-  
+
   object_class->finalize = hd_plugin_loader_python_finalize;
 
   loader_class->load = hd_plugin_loader_python_load;
